@@ -105,7 +105,8 @@ where the last three elements are optional."
     (mailing-list 4 ?,)
     (attribution 4 ?.)
     (organization 4)
-    (www 1)
+    (www 4)
+    (aka 4)
     (score -1)
     (x-face -1))
   "Alist of entry types for presentation.
@@ -132,6 +133,26 @@ entry cannot be modified."
   :group 'lsdb
   :type 'function)
 
+(defcustom lsdb-lookup-full-name-functions
+  '(lsdb-lookup-full-name-from-address-cache)
+  "List of functions to pick up the existing full-name of the sender.
+The sender is passed to each function as the argument."
+  :group 'lsdb
+  :type 'hook)
+
+(defcustom lsdb-update-record-functions
+  '(lsdb-update-address-cache)
+  "List of functions called after a record is updated.
+The updated record is passed to each function as the argument."
+  :group 'lsdb
+  :type 'hook)
+
+(defcustom lsdb-secondary-hash-tables
+  '(lsdb-address-cache)
+  "List of the hash tables for reverse lookup"
+  :group 'lsdb
+  :type 'list)
+
 (defcustom lsdb-window-max-height 7
   "Maximum number of lines used to display LSDB record."
   :group 'lsdb
@@ -227,11 +248,15 @@ Bourne shell or its equivalent \(not tcsh) is needed for \"2>\"."
 (defvar lsdb-hash-table nil
   "Internal hash table to hold LSDB records.")
 
+(defvar lsdb-address-cache nil
+  "The reverse lookup table for `lsdb-hash-table'.
+It represents address to full-name mapping.")
+
 (defvar lsdb-buffer-name "*LSDB*"
   "Buffer name to display LSDB record.")
 
-(defvar lsdb-hash-table-is-dirty nil
-  "Flag to indicate whether the hash table needs to be saved.")
+(defvar lsdb-hash-tables-are-dirty nil
+  "Flag to indicate whether the internal hash tables need to be saved.")
 
 (defvar lsdb-known-entry-names
   (make-vector 29 0)
@@ -296,44 +321,84 @@ This is the current number of slots in HASH-TABLE, whether occupied or not."
     (list 0 (make-vector (or (plist-get args :size) 29) 0))))
 
 ;;;_. Hash Table Reader/Writer
+(defconst lsdb-secondary-hash-table-start-format
+  ";;; %S\n")
+
+(defsubst lsdb-secondary-hash-table-start (hash-table)
+  (format lsdb-secondary-hash-table-start-format hash-table))
+
 (eval-and-compile
   (condition-case nil
       (progn
 	;; In XEmacs, hash tables can also be created by the lisp reader
 	;; using structure syntax.
 	(read-from-string "#s(hash-table)")
-	(defun lsdb-load-file (file)
-	  "Read the contents of FILE into a hash table."
-	  (let ((buffer (find-file-noselect file)))
-	    (unwind-protect
-		(save-excursion
-		  (set-buffer buffer)
-		  (re-search-forward "^#s")
-		  (beginning-of-line)
-		  (read (point-min-marker)))
-	      (kill-buffer buffer)))))
+	(defalias 'lsdb-read 'read))
     (invalid-read-syntax
-    (defun lsdb-load-file (file)
-      "Read the contents of FILE into a hash table."
-      (let* ((plist
-	      (with-temp-buffer
-		(insert-file-contents file)
-		(save-excursion
-		  (re-search-forward "^#s")
-		  (replace-match "")
-		  (beginning-of-line)
-		  (cdr (read (point-marker))))))
-	     (size (plist-get plist 'size))
-	     (data (plist-get plist 'data))
-	     (hash-table (lsdb-make-hash-table :size size :test 'equal)))
-	(while data
-	  (lsdb-puthash (pop data) (pop data) hash-table))
-	hash-table)))))
+     (defun lsdb-read (&optional marker)
+       "Read one Lisp expression as text from MARKER, return as Lisp object."
+       (save-excursion
+	 (goto-char marker)
+	 (if (looking-at "^#s(")
+	     (with-temp-buffer
+	       (buffer-disable-undo)
+	       (insert-buffer-substring (marker-buffer marker) marker)
+	       (goto-char (point-min))
+	       (delete-char 2)
+	       (let ((object (read (current-buffer)))
+		     hash-table data)
+		 (if (eq 'hash-table (car object))
+		     (progn
+		       (setq hash-table
+			     (lsdb-make-hash-table
+			      :size (plist-get (cdr object) 'size)
+			      :test 'equal)
+			     data (plist-get (cdr object) 'data))
+		       (while data
+			 (lsdb-puthash (pop data) (pop data) hash-table))
+		       hash-table)
+		   object)))))))))
 
-(defun lsdb-save-file (file hash-table)
-  "Write the entries within HASH-TABLE into FILE."
-  (let ((coding-system-for-write lsdb-file-coding-system))
-    (with-temp-file file
+(defun lsdb-load-hash-tables ()
+  "Read the contents of `lsdb-file' into the internal hash tables."
+  (let ((buffer (find-file-noselect lsdb-file))
+	tables)
+    (unwind-protect
+	(save-excursion
+	  (set-buffer buffer)
+	  (goto-char (point-min))
+	  (re-search-forward "^#s(")
+	  (goto-char (match-beginning 0))
+	  (setq lsdb-hash-table (lsdb-read (point-marker)))
+	  ;; Load the secondary hash tables following.
+	  (setq tables lsdb-secondary-hash-tables)
+	  (while tables
+	    (if (re-search-forward
+		 (concat "^" (lsdb-secondary-hash-table-start
+			      (car tables)))
+		 nil t)
+		(set (car tables) (lsdb-read (point-marker))))
+	    (setq tables (cdr tables))))
+      (kill-buffer buffer))))
+
+(defun lsdb-insert-hash-table (hash-table)
+  (insert "#s(hash-table size "
+	  ;; Reduce the actual size of the close hash table, because
+	  ;; XEmacs doesn't have a distinction between index-size and
+	  ;; hash-table-size.
+	  (number-to-string (lsdb-hash-table-count hash-table))
+	  " test equal data (")
+  (lsdb-maphash
+   (lambda (key value)
+     (insert (prin1-to-string key) " " (prin1-to-string value) " "))
+   hash-table)
+  (insert "))"))
+
+(defun lsdb-save-hash-tables ()
+  "Write the records within the internal hash tables into `lsdb-file'."
+  (let ((coding-system-for-write lsdb-file-coding-system)
+	tables)
+    (with-temp-file lsdb-file
       (if (and (or (featurep 'mule)
 		   (featurep 'file-coding))
 	       lsdb-file-coding-system)
@@ -346,15 +411,14 @@ This is the current number of slots in HASH-TABLE, whether occupied or not."
 				     lsdb-file-coding-system))))))
 	    (if coding-system-name
 		(insert ";;; -*- coding: " coding-system-name " -*-\n"))))
-      (insert "#s(hash-table size "
-	      ;; reduce the actual size of the close hash table
-	      (number-to-string (lsdb-hash-table-count hash-table))
-	      " test equal data (")
-      (lsdb-maphash
-       (lambda (key value)
-	 (insert (prin1-to-string key) " " (prin1-to-string value) " "))
-       hash-table)
-      (insert "))"))))
+      (lsdb-insert-hash-table lsdb-hash-table)
+      ;; Save the secondary hash tables following.
+      (setq tables lsdb-secondary-hash-tables)
+      (while tables
+	(insert "\n" (lsdb-secondary-hash-table-start
+		      (car tables)))
+	(lsdb-insert-hash-table (symbol-value (car tables)))
+	(setq tables (cdr tables))))))
 
 ;;;_. Mail Header Extraction
 (defun lsdb-fetch-field-bodies (regexp)
@@ -379,9 +443,9 @@ This is the current number of slots in HASH-TABLE, whether occupied or not."
   (let ((components (std11-extract-address-components string)))
     (if (nth 1 components)
 	(if (car components)
-	    (list (nth 1 components)
-		  (funcall lsdb-canonicalize-full-name-function
-			   (car components)))
+	    (list (funcall lsdb-canonicalize-full-name-function
+			   (car components))
+		  (nth 1 components))
 	  (list (nth 1 components) (nth 1 components))))))
 
 ;; stolen (and renamed) from nnheader.el
@@ -399,23 +463,96 @@ This is the current number of slots in HASH-TABLE, whether occupied or not."
       (set-buffer-multibyte multibyte))))
 
 ;;;_. Record Management
-(defun lsdb-maybe-load-file ()
+(defun lsdb-maybe-load-secondary-hash-tables ()
+  (let ((tables lsdb-secondary-hash-tables))
+    (while tables
+      (unless (symbol-value (car tables))
+	(set (car tables) (lsdb-make-hash-table :test 'equal))
+	(lsdb-maphash
+	 (lambda (key value)
+	   (run-hook-with-args
+	    'lsdb-update-record-functions
+	    (cons key value)))
+	 lsdb-hash-table)
+	(setq lsdb-hash-tables-are-dirty t))
+      (setq tables (cdr tables)))))
+
+(defun lsdb-maybe-load-hash-tables ()
   (unless lsdb-hash-table
     (if (file-exists-p lsdb-file)
-	(setq lsdb-hash-table (lsdb-load-file lsdb-file))
-      (setq lsdb-hash-table (lsdb-make-hash-table :test 'equal)))))
+	(lsdb-load-hash-tables)
+      (setq lsdb-hash-table (lsdb-make-hash-table :test 'equal)))
+    (lsdb-maybe-load-secondary-hash-tables)))
 
+;;;_ : Fallback Lookup Functions
+;;;_  , #1 Address Cache
+(defun lsdb-lookup-full-name-from-address-cache (sender)
+  (lsdb-gethash (nth 1 sender) lsdb-address-cache))
+
+(defun lsdb-update-address-cache (record)
+  (let ((net (cdr (assq 'net record))))
+    (while net
+      (lsdb-puthash (pop net) (car record) lsdb-address-cache))))
+
+;;;_  , #2 Iterate on the All Records (very slow)
+(defun lsdb-lookup-full-name-by-fuzzy-matching (sender)
+  (let ((names
+	 (if (string-match
+	      "\\`\\(.+\\)[ \t]+\\(/[ \t]+\\|(\\([^)]+\\))\\)"
+	      (car sender))
+	     (if (match-beginning 3)
+		 (list (match-string 1 (car sender))
+		       (match-string 3 (car sender)))
+	       (list (match-string 1 (car sender))
+		     (substring (car sender) (match-end 0))))
+	   (list (car sender))))
+	(case-fold-search t))
+    (catch 'found
+      (lsdb-maphash
+       (lambda (key value)
+	 (while names
+	   (if (or (string-match
+		    (concat "\\<" (regexp-quote (car names)) "\\>")
+		    key)
+		   (string-match
+		    (concat
+		     "\\<"
+		     (regexp-quote
+		      (mapconcat #'identity
+				 (nreverse (split-string (car names)))
+				 " "))
+		     "\\>")
+		    key)
+		   ;; Don't assume that we are using address cache.
+		   (member (nth 1 sender) (cdr (assq 'net value))))
+	       (throw 'found key))
+	   (setq names (cdr names))))
+       lsdb-hash-table))))
+
+;;;_ : Update Records
 (defun lsdb-update-record (sender &optional interesting)
-  (let ((old (lsdb-gethash (nth 1 sender) lsdb-hash-table))
-	(new (cons (cons 'net (list (car sender)))
+  (let ((old (lsdb-gethash (car sender) lsdb-hash-table))
+	(new (cons (cons 'net (list (nth 1 sender)))
 		   interesting))
 	merged
-	record)
+	record
+	full-name)
+    ;; Look for the existing record from the reverse hash table.
+    ;; If it is found, regsiter the current full-name as AKA.
+    (unless old
+      (setq full-name
+	    (run-hook-with-args-until-success
+	     'lsdb-lookup-full-name-functions
+	     sender))
+      (when full-name
+	(setq old (lsdb-gethash full-name lsdb-hash-table)
+	      new (cons (list 'aka (car sender)) new))
+	(setcar sender full-name)))
     (unless old
       (setq new (cons (cons 'creation-date (format-time-string "%Y-%m-%d"))
 		      new)))
     (setq merged (lsdb-merge-record-entries old new)
-	  record (cons (nth 1 sender) merged))
+	  record (cons (car sender) merged))
     (unless (equal merged old)
       (let ((entry (assq 'last-modified (cdr record)))
 	    (last-modified (format-time-string "%Y-%m-%d")))
@@ -425,11 +562,12 @@ This is the current number of slots in HASH-TABLE, whether occupied or not."
 			       (cdr record)))))
       (lsdb-puthash (car record) (cdr record)
 		    lsdb-hash-table)
-      (setq lsdb-hash-table-is-dirty t))
+      (run-hook-with-args 'lsdb-update-record-functions record)
+      (setq lsdb-hash-tables-are-dirty t))
     record))
 
 (defun lsdb-update-records ()
-  (lsdb-maybe-load-file)
+  (lsdb-maybe-load-hash-tables)
   (let (senders recipients interesting alist records bodies entry)
     (save-restriction
       (std11-narrow-to-header)
@@ -564,10 +702,33 @@ This is the current number of slots in HASH-TABLE, whether occupied or not."
 (defvar lsdb-last-candidates nil)
 (defvar lsdb-last-candidates-pointer nil)
 
+;;;_ : Matching Highlight
+(defvar lsdb-last-highlight-overlay nil)
+
+(defun lsdb-complete-name-highlight (start end)
+  (make-local-hook 'pre-command-hook)
+  (add-hook 'pre-command-hook 'lsdb-complete-name-highlight-update nil t)
+  (save-excursion
+    (goto-char start)
+    (search-forward lsdb-last-completion end)
+    (setq lsdb-last-highlight-overlay
+	  (make-overlay (match-beginning 0) (match-end 0)))
+    (overlay-put lsdb-last-highlight-overlay 'face
+		 (or (find-face 'isearch-secondary)
+		     'underline))))
+
+(defun lsdb-complete-name-highlight-update ()
+  (unless (eq 'this-command 'lsdb-complete-name)
+    (if lsdb-last-highlight-overlay
+	(delete-overlay lsdb-last-highlight-overlay))
+    (remove-hook 'pre-command-hook
+		 'lsdb-complete-name-highlight-update t)))
+
+;;;_ : Name Completion
 (defun lsdb-complete-name ()
   "Complete the user full-name or net-address before point"
   (interactive)
-  (lsdb-maybe-load-file)
+  (lsdb-maybe-load-hash-tables)
   (let* ((start
 	  (save-excursion
 	    (re-search-backward "\\(\\`\\|[\n:,]\\)[ \t]*")
@@ -580,7 +741,7 @@ This is the current number of slots in HASH-TABLE, whether occupied or not."
       (setq lsdb-last-candidates nil
 	    lsdb-last-candidates-pointer nil
 	    lsdb-last-completion (buffer-substring start (point))
-	    pattern (concat "\\<" lsdb-last-completion))
+	    pattern (concat "\\<" (regexp-quote lsdb-last-completion)))
       (lsdb-maphash
        (lambda (key value)
 	 (let ((net (cdr (assq 'net value))))
@@ -609,7 +770,8 @@ This is the current number of slots in HASH-TABLE, whether occupied or not."
       (setq lsdb-last-candidates-pointer lsdb-last-candidates))
     (when lsdb-last-candidates-pointer
       (delete-region start (point))
-      (insert (pop lsdb-last-candidates-pointer)))))
+      (insert (pop lsdb-last-candidates-pointer))
+      (lsdb-complete-name-highlight start (point)))))
 
 ;;;_. Major Mode (`lsdb-mode') Implementation
 ;;;_ : Modeline Buffer Identification
@@ -715,8 +877,8 @@ Modify whole identification by side effect."
       (font-lock-set-defaults)
     (set (make-local-variable 'font-lock-defaults)
 	 '(lsdb-font-lock-keywords t)))
-  (make-local-variable 'post-command-hook)
-  (setq post-command-hook 'lsdb-modeline-update)
+  (make-local-hook 'post-command-hook)
+  (add-hook 'post-command-hook 'lsdb-modeline-update nil t)
   (make-local-variable 'lsdb-modeline-string)
   (setq mode-line-buffer-identification
 	(lsdb-modeline-buffer-identification
@@ -794,7 +956,8 @@ the current record."
 	      (setcdr record (cons (cons ',entry-name form) (cdr record)))
 	      (lsdb-puthash (car record) (cdr record)
 			    lsdb-hash-table)
-	      (setq lsdb-hash-table-is-dirty t)
+	      (run-hook-with-args 'lsdb-update-record-functions record)
+	      (setq lsdb-hash-tables-are-dirty t)
 	      (beginning-of-line 2)
 	      (add-text-properties
 	       (point)
@@ -816,7 +979,8 @@ the current record."
       (setcdr record (delq entry (cdr record)))
       (lsdb-puthash (car record) (cdr record)
 		    lsdb-hash-table)
-      (setq lsdb-hash-table-is-dirty t))
+      (run-hook-with-args 'lsdb-update-record-functions record)
+      (setq lsdb-hash-tables-are-dirty t))
     (save-restriction
       (lsdb-narrow-to-record)
       (let ((case-fold-search t)
@@ -853,7 +1017,8 @@ the current record."
 		   (inhibit-read-only t)
 		   buffer-read-only)
 	      (setcdr entry form)
-	      (setq lsdb-hash-table-is-dirty t)
+	      (run-hook-with-args 'lsdb-update-record-functions record)
+	      (setq lsdb-hash-tables-are-dirty t)
 	      (lsdb-mode-delete-entry (symbol-name ',entry-name) t)
 	      (beginning-of-line)
 	      (add-text-properties
@@ -866,13 +1031,13 @@ the current record."
 (defun lsdb-mode-save (&optional dont-ask)
   "Save LSDB hash table into `lsdb-file'."
   (interactive)
-  (if (not lsdb-hash-table-is-dirty)
+  (if (not lsdb-hash-tables-are-dirty)
       (message "(No changes need to be saved)")
     (when (or (interactive-p)
 	      dont-ask
 	      (y-or-n-p "Save the LSDB now?"))
-      (lsdb-save-file lsdb-file lsdb-hash-table)
-      (setq lsdb-hash-table-is-dirty nil)
+      (lsdb-save-hash-tables)
+      (setq lsdb-hash-tables-are-dirty nil)
       (message "The LSDB was saved successfully."))))
 
 (defun lsdb-mode-quit-window (&optional kill window)
@@ -941,7 +1106,7 @@ performed against the entry field."
 	 "Search records regexp: ")
        nil nil nil 'lsdb-mode-lookup-history)
       entry-name)))
-  (lsdb-maybe-load-file)
+  (lsdb-maybe-load-hash-tables)
   (let ((records (lsdb-lookup-records regexp entry-name)))
     (if records
 	(lsdb-display-records records))))
@@ -1139,7 +1304,8 @@ of the buffer."
 				    (cdr (car records))))
 	(lsdb-puthash (car (car records)) (cdr (car records))
 		      lsdb-hash-table)
-	(setq lsdb-hash-table-is-dirty t)))))
+	(run-hook-with-args 'lsdb-update-record-functions (car records))
+	(setq lsdb-hash-tables-are-dirty t)))))
 
 (defun lsdb-mu-get-prefix-method ()
   "A mu-cite method to return a prefix from LSDB or \">\".
@@ -1229,11 +1395,15 @@ the user wants it."
 	       x-face)
       (goto-char (point-min))
       (end-of-line)
-      (put-text-property 0 1 'invisible t delimiter)
-      (put-text-property 0 (length delimiter) 'lsdb-record record delimiter)
-      (insert delimiter)
-      (while x-face
-	(funcall lsdb-insert-x-face-function (pop x-face))))))
+      (put-text-property 0 1 'invisible t delimiter) ;hide "\r"
+      (put-text-property
+       (point)
+       (progn
+	 (insert delimiter)
+	 (while x-face
+	   (funcall lsdb-insert-x-face-function (pop x-face)))
+	 (point))
+       'lsdb-record record))))
 
 (defun lsdb-insert-x-face-image (data type marker)
   (static-if (featurep 'xemacs)
@@ -1295,7 +1465,7 @@ the user wants it."
 (provide 'lsdb)
 
 (product-provide 'lsdb
-  (product-define "LSDB" nil '(0 2)))
+  (product-define "LSDB" nil '(0 4)))
 
 ;;;_* Local emacs vars.
 ;;; The following `outline-layout' local variable setting:
